((nil
  (indent-tabs-mode . nil))
 (emacs-lisp-mode
  (lisp-indent-local-overrides . ((cond . 0) (interactive . 0))))
 (makefile-mode
  (indent-tabs-mode . t))
 (git-commit-mode
  (git-commit-major-mode . git-commit-elisp-text-mode)))
